<!DOCTYPE kpartgui SYSTEM "kpartgui.dtd">

<kpartgui name="KStars" version="9">
<MenuBar noMerge="1">
        <Menu name="file" noMerge="1"><text>&amp;File</text>
                <Action name="new_window" />
                <Action name="close_window" />
                <Separator />
                <Action name="open_file" />
                <Action name="export_image" />
                <Action name="run_script" />
                <Separator />
                <Action name="printing_wizard" />
                <Action name="print" />
                <Separator />
                <Action name="quit"/>
        </Menu>

        <Menu name="time" noMerge="1"><text>T&amp;ime</text>
                <Action name="time_to_now" />
                <Action name="time_dialog" />
                <Separator />
                <Action name="time_step_backward" />
                <Action name="clock_startstop" />
                <Action name="time_step_forward" />
        </Menu>

        <Menu name="focus" noMerge="1"><text>&amp;Pointing</text>
                <Action name="zenith" />
                <Action name="north" />
                <Action name="east" />
                <Action name="south" />
                <Action name="west" />
                <Separator />
                <Action name="manual_focus" />
                <Separator />
                <Action name="find_object" />
                <Action name="track_object" />
        </Menu>

        <Menu name="view" noMerge="1"><text>&amp;View</text>
                <Action name="zoom_in" />
                <Action name="zoom_out" />
                <Action name="zoom_default" />
                <Action name="zoom_set" />
                <Separator />
                <Action name="fullscreen" />
                <Separator />
                <Action name="coordsys" />
                <Menu name="projection"><text>&amp;Projection</text>
                        <Action name="project_lambert"/>
                        <Action name="project_azequidistant"/>
                        <Action name="project_orthographic"/>
                        <Action name="project_equirectangular"/>
                        <Action name="project_stereographic"/>
                        <Action name="project_gnomonic"/>
                </Menu>
                <Action name="hipssources" />
        </Menu>

        <Menu name="tools" noMerge="1"><text>&amp;Tools</text>
                <Action name="astrocalculator" />
        <Menu name="devices" noMerge="1"><text>&amp;Devices</text>
                <!-- <Action name="telescope_wizard" /> -->
                <Action name="equipmentwriter"/>
                <Action name="device_manager" />
                <Action name="custom_drivers" />
                <Separator />
                <Action name="indi_cpl" />
        </Menu>

                <Action name="skycalendar" />
                <Action name="moonphasetool" />
                <Action name="altitude_vs_time" />
                <Action name="whats_up_tonight" />
                <Action name="whats_interesting" />
                <Action name="XPlanet" />
                <Action name="glossary" />
                <Action name="scriptbuilder" />
                <Action name="solarsystem" />
                <Action name="ekos" />
<!--                <Action name="jmoontool" /> -->
                <Action name="flagmanager" />
        </Menu>

        <Menu name="data" noMerge="1"><text>&amp;Data</text>
                <Action name="get_data" />
                <Action name="manual_add_dso" />
                <Menu name="updates"><text>&amp;Updates</text>
                <Action name="update_comets" />
                <Action name="update_asteroids" />
                <Action name="update_satellites" />
                <Action name="update_supernovae" />
                </Menu>
        </Menu>
        <Menu name="observation" noMerge="1"><text>&amp;Observation</text>
           <Action name="obslist" />
           <Action name="execute" />
           <Separator />
           <Action name="polaris_hour_angle" />
        </Menu>

        <Menu name="settings" noMerge="1"><text>&amp;Settings</text>
                <Menu name="infopanel"><text>&amp;Info Boxes</text>
                        <Action name="show_boxes"/>
                        <Separator />
                        <Action name="show_time_box" />
                        <Action name="show_focus_box" />
                        <Action name="show_location_box" />
                </Menu>
                <Merge name="StandardToolBarMenuHandler" />
                <Menu name="statusbar"><text>&amp;Statusbar</text>
                        <Action name="show_statusBar" />
                        <Separator />
                        <Action name="show_sbAzAlt" />
                        <Action name="show_sbRADec" />
                        <Action name="show_sbJ2000RADec" />
                </Menu>
                <Action name="themes" />
                <Action name="colorschemes" /> <!--This is a KMenuAction-->
                <Action name="fovsymbols" /> <!-- This is a KMenuAction-->
                <Action name="opengl" />
                <Action name="artificialhorizon" />
                <Separator />
                <Action name="manageobserver"/>
                <Action name="geolocation" />
                <Separator />
                <Action name="options_configure_keybinding"/>
                <Action name="options_configure_toolbars"/>
                <Action name="options_configure_notifications"/>
                <Action name="configure" />
                <Action name="startwizard" />
        </Menu>

        <Menu name="help"><text>&amp;Help</text>
                <Action name="help_contents"/>
                <Action name="help_tipofday"/>
                <Action name="help_whats_this"/>
                <Separator/>
                <Action name="help_report_bug"/>
                <Separator/>
                <Action name="switch_application_language"/>
                <Separator/>
                <Action name="help_about_app"/>
                <Action name="help_about_kde"/>
        </Menu>
</MenuBar>

<ToolBar noMerge="1" name="kstarsToolBar" iconText="icononly">
        <text>Main Toolbar</text>
        <Action name="zoom_in" />
        <Action name="zoom_out" />
        <Action name="find_object" />
        <Separator />
        <Action name="geolocation" />
        <Separator />
        <Action name="time_dialog" />
        <Action name="time_step_backward" />
        <Action name="clock_startstop" />
        <Action name="time_step_forward" />
        <Action name="timestep_control" />
        <Separator />
        <Action name="fovsymbols" /> <!-- This is a KMenuAction-->
        <Action name="hipssources" /> <!-- This is a KMenuAction-->
</ToolBar>

<ToolBar noMerge="1" name="viewToolBar" iconText="icononly">
        <text>View Toolbar</text>
        <Action name="show_stars" />
        <Action name="show_deepsky" />
        <Action name="show_planets" />
        <Action name="show_supernovae" />
        <Action name="show_satellites" />
        <Action name="show_clines" />
        <Action name="show_cnames" />
        <Action name="show_constellationart" />
        <Action name="show_cbounds" />
        <Action name="show_mw" />
        <Action name="show_equatorial_grid" />
        <Action name="show_horizontal_grid" />
        <Action name="show_horizon" />
        <Action name="show_whatsinteresting" />
</ToolBar>

<ToolBar noMerge="1" name="INDIToolBar" iconText="icononly">
        <text>INDI Toolbar</text>
        <Action name="show_ekos" />
        <Action name="show_control_panel" />
        <Action name="show_fits_viewer" />
        <Action name="show_sensor_fov" />
        <Action name="lock_telescope" />
</ToolBar>

<ToolBar noMerge="1" name="TelescopeToolBar" iconText="icononly" hidden="true">
        <text>Telescope Toolbar</text>
        <Action name="show_mount_box" />
        <Action name="telescope_track" />
        <Action name="telescope_slew" />
        <Action name="telescope_sync" />
        <Action name="telescope_park" />
        <Action name="telescope_unpark" />
        <Action name="telescope_abort" />
</ToolBar>

<ToolBar noMerge="1" name="DomeToolBar" iconText="icononly" hidden="true">
        <text>Dome Toolbar</text>
        <Action name="dome_park" />
        <Action name="dome_unpark" />
</ToolBar>

</kpartgui>
