<!DOCTYPE kpartgui SYSTEM "kpartgui.dtd">
<kpartgui name="FITSViewer" version="4">

<MenuBar noMerge="1">
<Menu name="file" noMerge="1"><text>&amp;File</text>
		<Action name="file_open" />
                <Action name="file_save"/>
                <Action name="file_save_as"/>
                <Separator/>
                <Action name="fits_header"/>
                <Action name="fits_debayer"/>
                <Separator/>
                <Action name="file_close"/>
</Menu>

<Menu name="edit"><text>&amp;Edit</text>
		<Action name="edit_undo"/>
		<Action name="edit_redo"/>
		<Separator/>
		<Action name="edit_copy"/>
</Menu>

<Menu name="view"><text>&amp;View</text>
		<Action name="view_zoom_in"/>
		<Action name="view_zoom_out"/>
		<Action name="view_actual_size"/>
                <Separator/>
                <Action name="filter1"/>
                <Action name="filter2"/>
                <Action name="filter3"/>
                <Action name="filter4"/>
                <Action name="filter5"/>
                <Action name="filter6"/>
                <Action name="filter7"/>
                <Action name="filter8"/>
                <Action name="filter9"/>
                <Action name="filter10"/>
                <Separator/>
                <Action name="mark_stars"/>
</Menu>

<Menu name="help"><text>&amp;Help</text>
                <Action name="help_contents"/>
                <Action name="help_whats_this"/>
                <Separator/>
                <Action name="help_report_bug"/>
                <Separator/>
                <Action name="switch_application_language"/>
                <Separator/>
                <Action name="help_about_app"/>
                <Action name="help_about_kde"/>
</Menu>

</MenuBar>

<ToolBar noMerge="1" name="mainToolBar" iconText="icononly"><text>Main Toolbar</text>
	<Action name="file_open" />
        <Action name="file_save" />
	<Action name="file_save_as" />
	<Action name="edit_undo" />
	<Action name="edit_redo" />
	<Action name="view_zoom_in" />
	<Action name="view_zoom_out" />	
	<Action name="view_actual_size"/>
        <Action name="view_zoom_fit"/>
        <Separator/>
        <Action name="rotate_right" />
        <Action name="rotate_left" />
        <Action name="flip_horizontal" />
        <Action name="flip_vertical" />
        <Separator/>
        <Action name="view_crosshair" />
        <Action name="view_pixel_grid" />
        <Action name="toggle_3D_graph" />
        <Separator/>
        <Action name="view_eq_grid" />
        <Action name="view_objects" />
        <Action name="center_telescope" />


        <Separator/>
</ToolBar>
	
<ToolBar noMerge="1" name="processToolBar" iconText="icononly"><text>Process ToolBar</text>
        <Separator/>
	<Action name="image_histogram" />
        <Action name="image_stats" />
        <Action name="image_stretch" />
</ToolBar>

</kpartgui>
