
#include "MeshIterator.h"

#include "HTMesh.h"
#include "MeshBuffer.h"

#include <iostream>

MeshIterator::MeshIterator(HTMesh *mesh, BufNum bufNum)
{
    cnt                = 0;
    MeshBuffer *buffer = mesh->meshBuffer(bufNum);
    m_size             = buffer->size();
    index              = buffer->buffer();
}
