xi18nc("string from libindi, used in the config dialog", "100x");
xi18nc("string from libindi, used in the config dialog", "10x");
xi18nc("string from libindi, used in the config dialog", "1200x");
xi18nc("string from libindi, used in the config dialog", "12x");
xi18nc("string from libindi, used in the config dialog", "600x");
xi18nc("string from libindi, used in the config dialog", "64x");
xi18nc("string from libindi, used in the config dialog", "900x");
xi18nc("string from libindi, used in the config dialog", ":CM#");
xi18nc("string from libindi, used in the config dialog", ":CMR#");
xi18nc("string from libindi, used in the config dialog", "AP UTC Offset");
xi18nc("string from libindi, used in the config dialog", "AP local time");
xi18nc("string from libindi, used in the config dialog", "AP sidereal time");
xi18nc("string from libindi, used in the config dialog", "Abell");
xi18nc("string from libindi, used in the config dialog", "Abort Motion");
xi18nc("string from libindi, used in the config dialog", "Abort Slew");
xi18nc("string from libindi, used in the config dialog", "Abort Slew/Track");
xi18nc("string from libindi, used in the config dialog", "Abort");
xi18nc("string from libindi, used in the config dialog", "Absolute Position");
xi18nc("string from libindi, used in the config dialog", "Activate");
xi18nc("string from libindi, used in the config dialog", "Active Filter");
xi18nc("string from libindi, used in the config dialog", "Actual Time");
xi18nc("string from libindi, used in the config dialog", "Alignment");
xi18nc("string from libindi, used in the config dialog", "Alt  D:M:S");
xi18nc("string from libindi, used in the config dialog", "Alt/Dec Anti-backlash");
xi18nc("string from libindi, used in the config dialog", "Alt/Dec PEC");
xi18nc("string from libindi, used in the config dialog", "AltAz");
xi18nc("string from libindi, used in the config dialog", "Anti Flicker");
xi18nc("string from libindi, used in the config dialog", "Aperture (mm)");
xi18nc("string from libindi, used in the config dialog", "Arp");
xi18nc("string from libindi, used in the config dialog", "Atmosphere");
xi18nc("string from libindi, used in the config dialog", "Auto");
xi18nc("string from libindi, used in the config dialog", "Auxiliary");
xi18nc("string from libindi, used in the config dialog", "Az D:M:S");
xi18nc("string from libindi, used in the config dialog", "Az/Ra Anti-backlash");
xi18nc("string from libindi, used in the config dialog", "Az/Ra PEC");
xi18nc("string from libindi, used in the config dialog", "Back Light");
xi18nc("string from libindi, used in the config dialog", "Backlash");
xi18nc("string from libindi, used in the config dialog", "Bias");
xi18nc("string from libindi, used in the config dialog", "Binning");
xi18nc("string from libindi, used in the config dialog", "Bits per pixel");
xi18nc("string from libindi, used in the config dialog", "Blue");
xi18nc("string from libindi, used in the config dialog", "Brightness");
xi18nc("string from libindi, used in the config dialog", "CCD Bias");
xi18nc("string from libindi, used in the config dialog", "CCD Information");
xi18nc("string from libindi, used in the config dialog", "CCD Maximum ADU");
xi18nc("string from libindi, used in the config dialog", "CCD Noise");
xi18nc("string from libindi, used in the config dialog", "CCD Simulator");
xi18nc("string from libindi, used in the config dialog", "CCD X Pixel Size");
xi18nc("string from libindi, used in the config dialog", "CCD X resolution");
xi18nc("string from libindi, used in the config dialog", "CCD Y Pixel Size");
xi18nc("string from libindi, used in the config dialog", "CCD Y resolution");
xi18nc("string from libindi, used in the config dialog", "CCD1");
xi18nc("string from libindi, used in the config dialog", "CCDs");
xi18nc("string from libindi, used in the config dialog", "Caldwell");
xi18nc("string from libindi, used in the config dialog", "Camera Model");
xi18nc("string from libindi, used in the config dialog", "Celsius");
xi18nc("string from libindi, used in the config dialog", "Centering");
xi18nc("string from libindi, used in the config dialog", "Color");
xi18nc("string from libindi, used in the config dialog", "Comet DEC motion arcmin/day");
xi18nc("string from libindi, used in the config dialog", "Comet RA motion arcmin/day");
xi18nc("string from libindi, used in the config dialog", "Comet tracking parameters");
xi18nc("string from libindi, used in the config dialog", "Comet");
xi18nc("string from libindi, used in the config dialog", "Communication");
xi18nc("string from libindi, used in the config dialog", "Compress");
xi18nc("string from libindi, used in the config dialog", "Compression");
xi18nc("string from libindi, used in the config dialog", "Configuration");
xi18nc("string from libindi, used in the config dialog", "Connect");
xi18nc("string from libindi, used in the config dialog", "Connection");
xi18nc("string from libindi, used in the config dialog", "Contrast");
xi18nc("string from libindi, used in the config dialog", "Control");
xi18nc("string from libindi, used in the config dialog", "Count");
xi18nc("string from libindi, used in the config dialog", "DEC (dd:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "Dark");
xi18nc("string from libindi, used in the config dialog", "Date");
xi18nc("string from libindi, used in the config dialog", "Date/Time");
xi18nc("string from libindi, used in the config dialog", "Date/Time/Location");
xi18nc("string from libindi, used in the config dialog", "Debug");
xi18nc("string from libindi, used in the config dialog", "Dec (arcmin)");
xi18nc("string from libindi, used in the config dialog", "Dec (dd:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "Dec D:M:S");
xi18nc("string from libindi, used in the config dialog", "Declination axis");
xi18nc("string from libindi, used in the config dialog", "Deep Sky Catalogs");
xi18nc("string from libindi, used in the config dialog", "Default");
xi18nc("string from libindi, used in the config dialog", "Device");
xi18nc("string from libindi, used in the config dialog", "Diff. Eq.");
xi18nc("string from libindi, used in the config dialog", "Direction");
xi18nc("string from libindi, used in the config dialog", "Disable");
xi18nc("string from libindi, used in the config dialog", "Disconnect");
xi18nc("string from libindi, used in the config dialog", "Dome control");
xi18nc("string from libindi, used in the config dialog", "Driver Info");
xi18nc("string from libindi, used in the config dialog", "Duration (s)");
xi18nc("string from libindi, used in the config dialog", "Duty cycle");
xi18nc("string from libindi, used in the config dialog", "EQ Coord");
xi18nc("string from libindi, used in the config dialog", "EQ PEC");
xi18nc("string from libindi, used in the config dialog", "East (msec)");
xi18nc("string from libindi, used in the config dialog", "East (sec)");
xi18nc("string from libindi, used in the config dialog", "East");
xi18nc("string from libindi, used in the config dialog", "East/West");
xi18nc("string from libindi, used in the config dialog", "Enable");
xi18nc("string from libindi, used in the config dialog", "Eq. Coordinates");
xi18nc("string from libindi, used in the config dialog", "Equatorial JNow");
xi18nc("string from libindi, used in the config dialog", "Exec");
xi18nc("string from libindi, used in the config dialog", "Expose");
xi18nc("string from libindi, used in the config dialog", "Expose Abort");
xi18nc("string from libindi, used in the config dialog", "Exposure");
xi18nc("string from libindi, used in the config dialog", "Extended GPS Features");
xi18nc("string from libindi, used in the config dialog", "Extrema");
xi18nc("string from libindi, used in the config dialog", "FWHM (arcseconds)");
xi18nc("string from libindi, used in the config dialog", "FWHM");
xi18nc("string from libindi, used in the config dialog", "Factory");
xi18nc("string from libindi, used in the config dialog", "Fan");
xi18nc("string from libindi, used in the config dialog", "Fast");
xi18nc("string from libindi, used in the config dialog", "Feed");
xi18nc("string from libindi, used in the config dialog", "Feedback");
xi18nc("string from libindi, used in the config dialog", "Field De-rotator");
xi18nc("string from libindi, used in the config dialog", "Filter #1");
xi18nc("string from libindi, used in the config dialog", "Filter #2");
xi18nc("string from libindi, used in the config dialog", "Filter #3");
xi18nc("string from libindi, used in the config dialog", "Filter #4");
xi18nc("string from libindi, used in the config dialog", "Filter #5");
xi18nc("string from libindi, used in the config dialog", "Filter Count");
xi18nc("string from libindi, used in the config dialog", "Filter Simulator");
xi18nc("string from libindi, used in the config dialog", "Filter Wheel");
xi18nc("string from libindi, used in the config dialog", "Filter");
xi18nc("string from libindi, used in the config dialog", "Find");
xi18nc("string from libindi, used in the config dialog", "Firmware Info");
xi18nc("string from libindi, used in the config dialog", "Firmware data");
xi18nc("string from libindi, used in the config dialog", "Firmware version");
xi18nc("string from libindi, used in the config dialog", "Firmware");
xi18nc("string from libindi, used in the config dialog", "Flat");
xi18nc("string from libindi, used in the config dialog", "Fluorescent");
xi18nc("string from libindi, used in the config dialog", "Focal Length (mm)");
xi18nc("string from libindi, used in the config dialog", "Focal Length (mm)");
xi18nc("string from libindi, used in the config dialog", "Focus Control");
xi18nc("string from libindi, used in the config dialog", "Focus In");
xi18nc("string from libindi, used in the config dialog", "Focus Out");
xi18nc("string from libindi, used in the config dialog", "Focus Speed");
xi18nc("string from libindi, used in the config dialog", "Focus Timer");
xi18nc("string from libindi, used in the config dialog", "Focus in");
xi18nc("string from libindi, used in the config dialog", "Focus out");
xi18nc("string from libindi, used in the config dialog", "Focuser Simulator");
xi18nc("string from libindi, used in the config dialog", "Focuser");
xi18nc("string from libindi, used in the config dialog", "Focusers");
xi18nc("string from libindi, used in the config dialog", "Frame Rate");
xi18nc("string from libindi, used in the config dialog", "Frame");
xi18nc("string from libindi, used in the config dialog", "FrameType");
xi18nc("string from libindi, used in the config dialog", "Freq");
xi18nc("string from libindi, used in the config dialog", "Full");
xi18nc("string from libindi, used in the config dialog", "GCVS");
xi18nc("string from libindi, used in the config dialog", "GOTO");
xi18nc("string from libindi, used in the config dialog", "GPS Power");
xi18nc("string from libindi, used in the config dialog", "GPS Status");
xi18nc("string from libindi, used in the config dialog", "GPS System");
xi18nc("string from libindi, used in the config dialog", "GPS/16 inch Features");
xi18nc("string from libindi, used in the config dialog", "Gamma");
xi18nc("string from libindi, used in the config dialog", "Generic Video4Linux");
xi18nc("string from libindi, used in the config dialog", "Geographic Location");
xi18nc("string from libindi, used in the config dialog", "Geographic Location");
xi18nc("string from libindi, used in the config dialog", "Goto");
xi18nc("string from libindi, used in the config dialog", "Green");
xi18nc("string from libindi, used in the config dialog", "Grey");
xi18nc("string from libindi, used in the config dialog", "Guide Abort");
xi18nc("string from libindi, used in the config dialog", "Guide East/West");
xi18nc("string from libindi, used in the config dialog", "Guide Head");
xi18nc("string from libindi, used in the config dialog", "Guide North/South");
xi18nc("string from libindi, used in the config dialog", "Guide West/East");
xi18nc("string from libindi, used in the config dialog", "Guide Wheel");
xi18nc("string from libindi, used in the config dialog", "Guide");
xi18nc("string from libindi, used in the config dialog", "Guider Aperture (mm)");
xi18nc("string from libindi, used in the config dialog", "Guider Control");
xi18nc("string from libindi, used in the config dialog", "Guider Focal Length (mm)");
xi18nc("string from libindi, used in the config dialog", "Guider Image");
xi18nc("string from libindi, used in the config dialog", "H Alpha");
xi18nc("string from libindi, used in the config dialog", "H:M:S");
xi18nc("string from libindi, used in the config dialog", "HA H:M:S");
xi18nc("string from libindi, used in the config dialog", "Halt");
xi18nc("string from libindi, used in the config dialog", "Height m");
xi18nc("string from libindi, used in the config dialog", "Height");
xi18nc("string from libindi, used in the config dialog", "High");
xi18nc("string from libindi, used in the config dialog", "Horizontal Coords");
xi18nc("string from libindi, used in the config dialog", "Hour axis");
xi18nc("string from libindi, used in the config dialog", "Hourangle Coords");
xi18nc("string from libindi, used in the config dialog", "Hue");
xi18nc("string from libindi, used in the config dialog", "Humidity Perc.");
xi18nc("string from libindi, used in the config dialog", "IC");
xi18nc("string from libindi, used in the config dialog", "Image Adjustments");
xi18nc("string from libindi, used in the config dialog", "Image Control");
xi18nc("string from libindi, used in the config dialog", "Image Data");
xi18nc("string from libindi, used in the config dialog", "Image Info");
xi18nc("string from libindi, used in the config dialog", "Image Settings");
xi18nc("string from libindi, used in the config dialog", "Image Type");
xi18nc("string from libindi, used in the config dialog", "Image");
xi18nc("string from libindi, used in the config dialog", "Indoor");
xi18nc("string from libindi, used in the config dialog", "Info");
xi18nc("string from libindi, used in the config dialog", "Jupiter");
xi18nc("string from libindi, used in the config dialog", "LX200 Basic");
xi18nc("string from libindi, used in the config dialog", "Lat (dd:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "Lat.  D:M:S +N");
xi18nc("string from libindi, used in the config dialog", "Lat. D:M:S +N");
xi18nc("string from libindi, used in the config dialog", "Left ");
xi18nc("string from libindi, used in the config dialog", "Library");
xi18nc("string from libindi, used in the config dialog", "Light");
xi18nc("string from libindi, used in the config dialog", "Limiting Mag");
xi18nc("string from libindi, used in the config dialog", "Load");
xi18nc("string from libindi, used in the config dialog", "Lon (dd:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "Long. D:M:S +E");
xi18nc("string from libindi, used in the config dialog", "Low");
xi18nc("string from libindi, used in the config dialog", "Luminosity");
xi18nc("string from libindi, used in the config dialog", "Lunar");
xi18nc("string from libindi, used in the config dialog", "Main Control");
xi18nc("string from libindi, used in the config dialog", "Manual Blue");
xi18nc("string from libindi, used in the config dialog", "Manual Red");
xi18nc("string from libindi, used in the config dialog", "Manual");
xi18nc("string from libindi, used in the config dialog", "Mars");
xi18nc("string from libindi, used in the config dialog", "Master alarm");
xi18nc("string from libindi, used in the config dialog", "Max slew Rate");
xi18nc("string from libindi, used in the config dialog", "Max");
xi18nc("string from libindi, used in the config dialog", "Max. travel");
xi18nc("string from libindi, used in the config dialog", "Maximum Tick");
xi18nc("string from libindi, used in the config dialog", "Maximum travel");
xi18nc("string from libindi, used in the config dialog", "Medium");
xi18nc("string from libindi, used in the config dialog", "Mercury");
xi18nc("string from libindi, used in the config dialog", "Messier");
xi18nc("string from libindi, used in the config dialog", "Minimum Tick");
xi18nc("string from libindi, used in the config dialog", "Mode");
xi18nc("string from libindi, used in the config dialog", "Model");
xi18nc("string from libindi, used in the config dialog", "Moon");
xi18nc("string from libindi, used in the config dialog", "Motion Control");
xi18nc("string from libindi, used in the config dialog", "Motion");
xi18nc("string from libindi, used in the config dialog", "Motor Steps");
xi18nc("string from libindi, used in the config dialog", "Motor steps per tick");
xi18nc("string from libindi, used in the config dialog", "Mount coordinates");
xi18nc("string from libindi, used in the config dialog", "Mount init.");
xi18nc("string from libindi, used in the config dialog", "Mounting");
xi18nc("string from libindi, used in the config dialog", "Move to rate");
xi18nc("string from libindi, used in the config dialog", "Movement Control");
xi18nc("string from libindi, used in the config dialog", "NGC");
xi18nc("string from libindi, used in the config dialog", "Name");
xi18nc("string from libindi, used in the config dialog", "Neptune");
xi18nc("string from libindi, used in the config dialog", "Noise Reduction");
xi18nc("string from libindi, used in the config dialog", "None");
xi18nc("string from libindi, used in the config dialog", "North (msec)");
xi18nc("string from libindi, used in the config dialog", "North (sec)");
xi18nc("string from libindi, used in the config dialog", "North");
xi18nc("string from libindi, used in the config dialog", "North/South");
xi18nc("string from libindi, used in the config dialog", "Note");
xi18nc("string from libindi, used in the config dialog", "Number");
xi18nc("string from libindi, used in the config dialog", "OFF");
xi18nc("string from libindi, used in the config dialog", "ON");
xi18nc("string from libindi, used in the config dialog", "OTA Temperature (C)");
xi18nc("string from libindi, used in the config dialog", "OTA Update");
xi18nc("string from libindi, used in the config dialog", "Oag Offset (arcminutes)");
xi18nc("string from libindi, used in the config dialog", "Object Info");
xi18nc("string from libindi, used in the config dialog", "Object Number");
xi18nc("string from libindi, used in the config dialog", "Object");
xi18nc("string from libindi, used in the config dialog", "Off");
xi18nc("string from libindi, used in the config dialog", "Offset");
xi18nc("string from libindi, used in the config dialog", "On Set");
xi18nc("string from libindi, used in the config dialog", "On");
xi18nc("string from libindi, used in the config dialog", "Options");
xi18nc("string from libindi, used in the config dialog", "Outdoor");
xi18nc("string from libindi, used in the config dialog", "PE N/S");
xi18nc("string from libindi, used in the config dialog", "PE W/E");
xi18nc("string from libindi, used in the config dialog", "Park Scope");
xi18nc("string from libindi, used in the config dialog", "Park");
xi18nc("string from libindi, used in the config dialog", "Periodic Error");
xi18nc("string from libindi, used in the config dialog", "Philips Webcam");
xi18nc("string from libindi, used in the config dialog", "Pixel size (um)");
xi18nc("string from libindi, used in the config dialog", "Pixel size X");
xi18nc("string from libindi, used in the config dialog", "Pixel size Y");
xi18nc("string from libindi, used in the config dialog", "Pluto");
xi18nc("string from libindi, used in the config dialog", "Polar");
xi18nc("string from libindi, used in the config dialog", "Port");
xi18nc("string from libindi, used in the config dialog", "Ports");
xi18nc("string from libindi, used in the config dialog", "Position");
xi18nc("string from libindi, used in the config dialog", "Power");
xi18nc("string from libindi, used in the config dialog", "Pressure hPa");
xi18nc("string from libindi, used in the config dialog", "Property");
xi18nc("string from libindi, used in the config dialog", "RA  H:M:S");
xi18nc("string from libindi, used in the config dialog", "RA (arcmin)");
xi18nc("string from libindi, used in the config dialog", "RA (hh:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "RA H:M:S");
xi18nc("string from libindi, used in the config dialog", "RA motor");
xi18nc("string from libindi, used in the config dialog", "Ra (hh:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "Rate");
xi18nc("string from libindi, used in the config dialog", "Raw");
xi18nc("string from libindi, used in the config dialog", "Red");
xi18nc("string from libindi, used in the config dialog", "Relative Position");
xi18nc("string from libindi, used in the config dialog", "Resolution x");
xi18nc("string from libindi, used in the config dialog", "Resolution y");
xi18nc("string from libindi, used in the config dialog", "Restart");
xi18nc("string from libindi, used in the config dialog", "Restore");
xi18nc("string from libindi, used in the config dialog", "SAO");
xi18nc("string from libindi, used in the config dialog", "STAR");
xi18nc("string from libindi, used in the config dialog", "Saturation Mag");
xi18nc("string from libindi, used in the config dialog", "Saturn");
xi18nc("string from libindi, used in the config dialog", "Save home");
xi18nc("string from libindi, used in the config dialog", "Save");
xi18nc("string from libindi, used in the config dialog", "Scope Location");
xi18nc("string from libindi, used in the config dialog", "Scope Properties");
xi18nc("string from libindi, used in the config dialog", "Seeing");
xi18nc("string from libindi, used in the config dialog", "Select item...");
xi18nc("string from libindi, used in the config dialog", "Select");
xi18nc("string from libindi, used in the config dialog", "Selenographic Sync");
xi18nc("string from libindi, used in the config dialog", "Set Register");
xi18nc("string from libindi, used in the config dialog", "Set home");
xi18nc("string from libindi, used in the config dialog", "Set register");
xi18nc("string from libindi, used in the config dialog", "Set");
xi18nc("string from libindi, used in the config dialog", "Settings");
xi18nc("string from libindi, used in the config dialog", "Shutter Speed");
xi18nc("string from libindi, used in the config dialog", "Sidereal");
xi18nc("string from libindi, used in the config dialog", "Sidereal Time");
xi18nc("string from libindi, used in the config dialog", "Sidereal time");
xi18nc("string from libindi, used in the config dialog", "Simulation");
xi18nc("string from libindi, used in the config dialog", "Simulator Config");
xi18nc("string from libindi, used in the config dialog", "Simulator Settings");
xi18nc("string from libindi, used in the config dialog", "Site 1");
xi18nc("string from libindi, used in the config dialog", "Site 2");
xi18nc("string from libindi, used in the config dialog", "Site 3");
xi18nc("string from libindi, used in the config dialog", "Site 4");
xi18nc("string from libindi, used in the config dialog", "Site Management");
xi18nc("string from libindi, used in the config dialog", "Site Name");
xi18nc("string from libindi, used in the config dialog", "Sites");
xi18nc("string from libindi, used in the config dialog", "Sky Glow (magnitudes)");
xi18nc("string from libindi, used in the config dialog", "Sleep");
xi18nc("string from libindi, used in the config dialog", "Slew Accuracy");
xi18nc("string from libindi, used in the config dialog", "Slew rate");
xi18nc("string from libindi, used in the config dialog", "Slew");
xi18nc("string from libindi, used in the config dialog", "Slow");
xi18nc("string from libindi, used in the config dialog", "Snoop dc connection");
xi18nc("string from libindi, used in the config dialog", "Snoop dc master alarm");
xi18nc("string from libindi, used in the config dialog", "Snoop dc mode");
xi18nc("string from libindi, used in the config dialog", "Snoop devices");
xi18nc("string from libindi, used in the config dialog", "Solar System");
xi18nc("string from libindi, used in the config dialog", "Solar");
xi18nc("string from libindi, used in the config dialog", "South (msec)");
xi18nc("string from libindi, used in the config dialog", "South (sec)");
xi18nc("string from libindi, used in the config dialog", "South");
xi18nc("string from libindi, used in the config dialog", "Speed");
xi18nc("string from libindi, used in the config dialog", "Star Catalogs");
xi18nc("string from libindi, used in the config dialog", "Step delay");
xi18nc("string from libindi, used in the config dialog", "Stop");
xi18nc("string from libindi, used in the config dialog", "Stream Off");
xi18nc("string from libindi, used in the config dialog", "Stream On");
xi18nc("string from libindi, used in the config dialog", "Swap buttons");
xi18nc("string from libindi, used in the config dialog", "Switch 1");
xi18nc("string from libindi, used in the config dialog", "Switch 2");
xi18nc("string from libindi, used in the config dialog", "Switch 3");
xi18nc("string from libindi, used in the config dialog", "Switch 4");
xi18nc("string from libindi, used in the config dialog", "Sync");
xi18nc("string from libindi, used in the config dialog", "Telescope Simulator");
xi18nc("string from libindi, used in the config dialog", "Telescope");
xi18nc("string from libindi, used in the config dialog", "Telescopes");
xi18nc("string from libindi, used in the config dialog", "Temma Driver");
xi18nc("string from libindi, used in the config dialog", "Temma version");
xi18nc("string from libindi, used in the config dialog", "Temma");
xi18nc("string from libindi, used in the config dialog", "Temp.");
xi18nc("string from libindi, used in the config dialog", "Temperature K");
xi18nc("string from libindi, used in the config dialog", "Theta D:M:S");
xi18nc("string from libindi, used in the config dialog", "Ticks");
xi18nc("string from libindi, used in the config dialog", "Time Factor");
xi18nc("string from libindi, used in the config dialog", "Time");
xi18nc("string from libindi, used in the config dialog", "Timer (ms)");
xi18nc("string from libindi, used in the config dialog", "Timer");
xi18nc("string from libindi, used in the config dialog", "Top");
xi18nc("string from libindi, used in the config dialog", "Total Exposure Time (ms)");
xi18nc("string from libindi, used in the config dialog", "Track Mode");
xi18nc("string from libindi, used in the config dialog", "Track");
xi18nc("string from libindi, used in the config dialog", "Tracking Accuracy");
xi18nc("string from libindi, used in the config dialog", "Tracking Frequency");
xi18nc("string from libindi, used in the config dialog", "Tracking Mode");
xi18nc("string from libindi, used in the config dialog", "Tracking mode");
xi18nc("string from libindi, used in the config dialog", "Transformation");
xi18nc("string from libindi, used in the config dialog", "UGC");
xi18nc("string from libindi, used in the config dialog", "UTC Offset");
xi18nc("string from libindi, used in the config dialog", "UTC Time");
xi18nc("string from libindi, used in the config dialog", "UTC");
xi18nc("string from libindi, used in the config dialog", "Update Client");
xi18nc("string from libindi, used in the config dialog", "Update GPS");
xi18nc("string from libindi, used in the config dialog", "Update");
xi18nc("string from libindi, used in the config dialog", "Uranus");
xi18nc("string from libindi, used in the config dialog", "Use Pulse Cmd");
xi18nc("string from libindi, used in the config dialog", "Venus");
xi18nc("string from libindi, used in the config dialog", "Version");
xi18nc("string from libindi, used in the config dialog", "Video Stream");
xi18nc("string from libindi, used in the config dialog", "Video");
xi18nc("string from libindi, used in the config dialog", "Wake up");
xi18nc("string from libindi, used in the config dialog", "Webcam Name");
xi18nc("string from libindi, used in the config dialog", "West (msec)");
xi18nc("string from libindi, used in the config dialog", "West (sec)");
xi18nc("string from libindi, used in the config dialog", "West");
xi18nc("string from libindi, used in the config dialog", "West/East");
xi18nc("string from libindi, used in the config dialog", "W/E Rate");
xi18nc("string from libindi, used in the config dialog", "N/S Rate");
xi18nc("string from libindi, used in the config dialog", "White Balance Mode");
xi18nc("string from libindi, used in the config dialog", "White Balance");
xi18nc("string from libindi, used in the config dialog", "Whiteness");
xi18nc("string from libindi, used in the config dialog", "Width");
xi18nc("string from libindi, used in the config dialog", "X");
xi18nc("string from libindi, used in the config dialog", "Y");
xi18nc("string from libindi, used in the config dialog", "app. to refracted");
xi18nc("string from libindi, used in the config dialog", "app., refr., tel., observed");
xi18nc("string from libindi, used in the config dialog", "app., refr., telescope");
xi18nc("string from libindi, used in the config dialog", "arcseconds");
xi18nc("string from libindi, used in the config dialog", "cold");
xi18nc("string from libindi, used in the config dialog", "danger");
xi18nc("string from libindi, used in the config dialog", "decPEC (dd:mm:ss)");
xi18nc("string from libindi, used in the config dialog", "dome control");
xi18nc("string from libindi, used in the config dialog", "identity");
xi18nc("string from libindi, used in the config dialog", "lunar");
xi18nc("string from libindi, used in the config dialog", "manual");
xi18nc("string from libindi, used in the config dialog", "max Alt");
xi18nc("string from libindi, used in the config dialog", "min Alt");
xi18nc("string from libindi, used in the config dialog", "off");
xi18nc("string from libindi, used in the config dialog", "on");
xi18nc("string from libindi, used in the config dialog", "rel. to HA");
xi18nc("string from libindi, used in the config dialog", "reset");
xi18nc("string from libindi, used in the config dialog", "sidereal");
xi18nc("string from libindi, used in the config dialog", "solar");
xi18nc("string from libindi, used in the config dialog", "undefined");
xi18nc("string from libindi, used in the config dialog", "warm");
