/*
  Copyright (C) 2015-2017, Pavel Mraz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#pragma once

#include "hips.h"

#include <QCache>

class PixCache
{
public:
  PixCache() = default;

  void add(pixCacheKey_t &key, pixCacheItem_t *item, int cost);
  pixCacheItem_t *get(pixCacheKey_t &key);
  void setMaxCost(int maxCost);
  void printCache();
  int  used();

private:  
  QCache <pixCacheKey_t, pixCacheItem_t> m_cache;
};

